#!/usr/bin/python3

# Copyright (c) 2020-2022, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import os
import sys
import argparse
import logging
import urllib.request
import pathlib
import re

logger = logging.getLogger()
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)

iGPU_package = 'nvidia-l4t-cuda'
dGPU_package = 'cuda-drivers'
l4t_repo = 'repo.download.nvidia.com/jetson'
required_kernel = '5.10.65-tegra'

def get_package_version(package_name):
    """ Returns the version of a package installed on the system, or None if not installed. """
    dpkg_output = os.popen(f"dpkg -l {package_name} 2> /dev/null").read().splitlines()
    if len(dpkg_output):
        package_details = dpkg_output[-1].split()
        if package_details[0] == 'ii':
            return package_details[2]
        else:
            return None
    else:
        return None

def get_cuda_dependent_packages(cuda_version):
    """ Returns the list of installed packages that depend on CUDA (based on their package version). """
    packages = []
    dpkg_output = os.popen(f"dpkg -l").read().splitlines()
    for line in dpkg_output:
        parts = line.split()
        if parts[0] != 'ii':
            continue
        version = parts[2]
        if cuda_version in version:
            packages.append(parts[1])
    return packages

def get_repo_packages(repo_name):
    """ Returns the list of unique package names that are provided by an apt repo. """
    packages = []
    repo_name = repo_name.replace('/', '_')
    file_name = None
    for root, dirs, files in os.walk('/var/lib/apt/lists'):
        for name in files:
            if re.search(f"{repo_name}.*_Packages", name):
                file_name = f"{root}/{name}"
                break
    if file_name:
        with open(file_name) as package_file:
            lines = package_file.read().splitlines()
            for line in lines:
                match = re.match('Package: (.*)', line)
                if match and match.group(1) not in packages:
                    packages.append(match.group(1))
    return packages

def get_installed_repo_packages(repo_name):
    """ Returns the list packages that are installed by an apt repo. """
    installed_packages = []
    packages = get_repo_packages(repo_name)
    for package in packages:
        if get_package_version(package):
            installed_packages.append(package)
    return installed_packages

def get_soc():
    """ Returns the SOC detected ('t194' or 't234' string). """
    soc = os.popen("cat /etc/nv_boot_control.conf | grep TEGRA_CHIPID | cut -d ' ' -f 2").read().strip()
    if soc == "0x19":
        return "t194"
    elif soc == "0x23":
        return "t234"
    else:
        logger.error(f"Unknown SOC type, unable to determine SOC repo location.")
        sys.exit(1)

def get_current_driver():
    """ Returns the currently installed driver ('iGPU' or 'dGPU' string). """
    if get_package_version(iGPU_package):
        return 'iGPU'
    elif get_package_version(dGPU_package):
        return 'dGPU'
    else:
        return None

def add_command(command_list, description, command):
    """ Adds a new command to a command list. """
    command_list.append({'description': description, 'command': command})

def print_commands(command_list, verbose):
    """ Prints the contents of a command list. """
    command_number = 0
    for command in command_list:
        command_number += 1
        if verbose:
            logger.info(f"[{command_number}/{len(command_list)}] {command['description']}:")
            logger.info(f"{command['command']}")
        else:
            logger.info(f"[{command_number}/{len(command_list)}] {command['command']}")

def run_commands(command_list, interactive, logfile):
    """ Run the contents of a command list. """
    if os.geteuid() != 0:
        logger.error("Installing drivers must be done with root. Please use 'sudo' to run script.")
        sys.exit(1)
    try:
        log_suffix = ""
        if logfile:
            log_suffix = f" 2>&1 | tee -a {logfile}"
            assert os.system(f"rm -f {logfile}") == 0
        assert os.system(f"echo '=== STARTING INSTALL ===\n'{log_suffix}") == 0
        command_number = 0
        for command in command_list:
            command_number += 1
            skip = False
            while interactive:
                logger.info(f"[{command_number}/{len(command_list)}] {command['description']}:")
                logger.info(f"  {command['command']}")
                user_input = input("Continue? (Yes/No/Skip) [Y/n/s]")
                if user_input == 'N' or user_input == 'n':
                    logger.info("Terminating installation.")
                    sys.exit(130)
                elif user_input == 'S' or user_input == 's':
                    skip = True
                    break
                elif user_input == 'Y' or user_input == 'y' or user_input == '':
                    break
            command_log = f"\"[{command_number}/{len(command_list)}] "
            if skip:
                command_log += "Skipping.\n"
            else:
                command_log += "Executing.\n"
            command_log += f"# {command['command']}\n\""
            assert os.system(f"echo {command_log}{log_suffix}") == 0
            if not skip:
                command_str = command['command'].replace(" &&", f"{log_suffix} &&") + log_suffix
                assert os.system(command_str) == 0, f"ERROR: {command['description']} failed!"
        assert os.system(f"echo '=== INSTALL COMPLETE ==='{log_suffix}") == 0
    except AssertionError as error:
        logger.error(error)
        sys.exit(1)
    except KeyboardInterrupt as error:
        sys.exit(130)
    except Exception as ex:
        logger.error("An exception of type {0} occurred. Arguments:\n{1!r}".format(type(ex).__name__, ex.args))
        sys.exit(1)

def remove_igpu(commands):
    """ Add commands to remove the iGPU drivers and apt repos. """
    packages_to_remove = get_installed_repo_packages("common")
    packages_to_remove += get_installed_repo_packages(get_soc())
    packages_to_remove.sort()
    # If the iGPU repos do not exist, then the packages have likely been installed by
    # SDK Manager and this is the first time running this script. In this case, remove
    # all nvidia-l4t and cuda-related packages.
    if len(packages_to_remove) == 0:
        print ("no pkg info for igpu")
        packages_to_remove = ["cuda-*"]
        dpkg_output = os.popen(f"dpkg -l").read().splitlines()
        for line in dpkg_output:
            parts = line.split()
            if parts[0] != 'ii':
                continue
            if "nvidia-l4t-" in parts[1]:
                packages_to_remove += [parts[1]]
        packages_to_remove += get_cuda_dependent_packages("cuda11.4")
    packages_to_remove.remove("nvidia-l4t-core")
    packages_to_remove.remove("nvidia-l4t-gputools")
    remove_string = " ".join(packages_to_remove)
    add_command(commands, "Remove iGPU driver packages",
        f"apt purge -y {remove_string} && apt autoremove -y")
    add_command(commands, "Remove stale library config",
        "rm -f /etc/ld.so.conf.d/aarch64-linux-gnu_EGL.conf")
    add_command(commands, "Remove iGPU apt repos",
        "rm -f /etc/apt/sources.list.d/l4t.list && apt update")
    add_command(commands, "Remove iGPU nvpmodel configuration",
        "rm -f /etc/nvpmodel.conf /var/lib/nvpmodel/status")

def remove_dgpu(commands):
    """ Add commands to remove the dGPU drivers and apt repos. """
    packages_to_remove = get_installed_repo_packages("dgpu-rm")
    packages_to_remove += get_installed_repo_packages("libnvidia-container/stable/ubuntu18.04")
    packages_to_remove += get_installed_repo_packages("nvidia-container-runtime/stable/ubuntu18.04")
    packages_to_remove.remove("nvidia-l4t-core")
    packages_to_remove.remove("nvidia-l4t-gputools")
    packages_to_remove.sort()
    remove_string = " ".join(packages_to_remove)
    add_command(commands, "Remove dGPU driver packages",
        f"apt purge -y {remove_string} && apt autoremove -y")
    add_command(commands, "Remove dGPU apt repos",
        "rm -f /etc/apt/sources.list.d/l4t_rm.list && apt update")
    add_command(commands, "Remove nvidia-container-runtime config",
        "rm -rf /etc/systemd/system/docker.service.d")
    add_command(commands, "Remove dGPU nvpmodel configuration",
        "rm -f /etc/nvpmodel.conf /var/lib/nvpmodel/status")

def install_igpu(commands, l4t_repo):
    """ Add commands to install the iGPU driver packages. """
    add_command(commands, "Remove blacklist for iGPU driver",
        "rm -f /etc/modprobe.d/blacklist-nvgpu.conf")
    add_command(commands, "Add blacklist for dGPU driver",
        "echo 'blacklist nvidia' > /etc/modprobe.d/blacklist-nvidia.conf && " +
        "echo 'blacklist nvidia_modeset' >> /etc/modprobe.d/blacklist-nvidia.conf && " +
        "echo 'blacklist nvidia_drm' >> /etc/modprobe.d/blacklist-nvidia.conf && " +
        "echo 'blacklist nvidia_uvm' >> /etc/modprobe.d/blacklist-nvidia.conf")
    add_command(commands, "Add public key for iGPU L4T apt repo",
        f"apt-key adv --fetch-keys http://{l4t_repo}/jetson-ota-public.asc")
    add_command(commands, "Add iGPU L4T apt repos",
        f"echo 'deb http://{l4t_repo}/{get_soc()} r34.1 main' > /etc/apt/sources.list.d/l4t.list && " +
        f"echo 'deb http://{l4t_repo}/common r34.1 main' >> /etc/apt/sources.list.d/l4t.list")
    add_command(commands, "Install iGPU drivers",
        "apt update && apt install -y nvidia-l4t-* nvidia-jetpack")
    if l4t_repo == globals()['l4t_repo']:
        add_command(commands, "Remove iGPU apt repos",
            "rm -f /etc/apt/sources.list.d/l4t.list && apt update")
    add_command(commands, "Install iGPU nvpmodel configuration",
        "ln -sf /etc/nvpmodel/nvpmodel_t194_e3900_iGPU.conf /etc/nvpmodel.conf")

def install_dgpu(commands, l4t_repo):
    """ Add commands to install the dGPU driver packages. """
    packages_to_install = [
        "nvidia-l4t-*",
        "nvidia-driver-510",
        "nvidia-dkms-510",
        "nvidia-utils-510",
        "cuda",
        "nvidia-container-runtime",
        "libnvinfer-bin"
    ]
    install_string = " ".join(packages_to_install)
    add_command(commands, "Remove blacklist for dGPU driver",
        "rm -f /etc/modprobe.d/blacklist-nvidia.conf")
    add_command(commands, "Add blacklist for iGPU driver",
        "echo 'blacklist nvgpu' > /etc/modprobe.d/blacklist-nvgpu.conf")
    add_command(commands, "Add public key for dGPU L4T apt repo",
        f"apt-key adv --fetch-keys http://{l4t_repo}/jetson-ota-public.asc")
    add_command(commands, "Add dGPU L4T apt repo",
        f"echo 'deb http://{l4t_repo}/dgpu-rm r34.1 main'" +
        " >> /etc/apt/sources.list.d/l4t_rm.list")
    add_command(commands, "Add public key for nvidia-container-runtime apt repo",
        "apt-key adv --fetch-keys https://nvidia.github.io/nvidia-container-runtime/gpgkey")
    add_command(commands, "Add nvidia-container-runtime apt repo",
        f"echo 'deb https://nvidia.github.io/libnvidia-container/stable/ubuntu18.04/$(ARCH) /'" +
         " >> /etc/apt/sources.list.d/l4t_rm.list && " +
        f"echo 'deb https://nvidia.github.io/nvidia-container-runtime/stable/ubuntu18.04/$(ARCH) /'" +
         " >> /etc/apt/sources.list.d/l4t_rm.list")
    add_command(commands, "Install dGPU drivers",
        f"apt update && apt install -y {install_string}")
    add_command(commands, "Repair dGPU library config",
        "echo '/usr/lib/aarch64-linux-gnu/tegra' >> /etc/ld.so.conf.d/nvidia-tegra.conf && ldconfig")
    add_command(commands, "Configure nvidia-container-runtime",
        "mkdir /etc/systemd/system/docker.service.d && " +
        "echo '[Service]' > /etc/systemd/system/docker.service.d/override.conf && " +
        "echo 'ExecStart=' >> /etc/systemd/system/docker.service.d/override.conf && " +
        "echo 'ExecStart=/usr/bin/dockerd --host=fd:// --add-runtime=nvidia=/usr/bin/nvidia-container-runtime'" +
        " >> /etc/systemd/system/docker.service.d/override.conf")
    add_command(commands, "Install dGPU nvpmodel configuration",
        "ln -sf /etc/nvpmodel/nvpmodel_t194_e3900_dGPU.conf /etc/nvpmodel.conf")

def install_apt_pin_file(commands, l4t_repo):
    """ Install APT pin file from ${l4t_repo} """
    apt_pin_filename = 'jetson-clara-pin-600'

    local_apt_pin_file = f"/etc/apt/preferences.d/{apt_pin_filename}"
    if os.path.exists(local_apt_pin_file):
        add_command(commands, "Remove existing APT pin file", f"rm {local_apt_pin_file}")

    remote_apt_pin_file = f"{l4t_repo}/{apt_pin_filename}"
    add_command(commands, f"Install APT pin file from {l4t_repo}",
            f"wget {remote_apt_pin_file} -P /etc/apt/preferences.d")

def query():
    """ Displays details of the currently installed driver stack. """
    iGPU_version = get_package_version(iGPU_package)
    dGPU_version = get_package_version(dGPU_package)
    if iGPU_version:
        logger.info(f"iGPU ({iGPU_package}, {iGPU_version})")
    elif dGPU_version:
        logger.info(f"dGPU ({dGPU_package}, {dGPU_version})")
        nvidia_smi_output = os.popen("nvidia-smi --query-gpu=gpu_name,memory.total --format=csv").read().splitlines()
        if len(nvidia_smi_output):
            logger.info(nvidia_smi_output[-1])
    else:
        logger.info("No drivers currently loaded.")

def install(args):
    """ Installs the iGPU or dGPU driver stack. """
    current_driver = get_current_driver()
    new_driver = args['driver'][0]

    if current_driver == new_driver and not args['force']:
        logger.info(f"{new_driver} drivers already installed; exiting.")
        sys.exit(0)

    # Ensure that the kernel version matches the expected version.
    kernel_version = os.popen("uname -r").read().strip()
    if kernel_version != required_kernel:
        logger.error(f"ERROR: kernel version ({kernel_version}) does not match required version ({required_kernel}).")
        logger.error("Please ensure that the device has been flashed with Holoscan SDK.")
        sys.exit(1)

    # Ensure that the apt repo is accessible.
    if not args['dry']:
        try:
            repo_status_code = urllib.request.urlopen(f"http://{args['l4t_repo']}/jetson-ota-public.asc").getcode()
        except urllib.request.URLError as e:
            logger.error(f"ERROR: L4T repo address (http://{args['l4t_repo']}) is not reachable.")
            logger.error("Please check that you are connected to the internet.")
            sys.exit(1)

    commands = []

    logger.info(f"Preparing commands to install {new_driver}. This may take a few moments.")

    install_apt_pin_file(commands, args['l4t_repo'])

    # Add commands to remove the current driver.
    if current_driver == 'iGPU':
        remove_igpu(commands)
    elif current_driver == 'dGPU':
        remove_dgpu(commands)

    # Add commands to install the new driver.
    if new_driver == 'iGPU':
        install_igpu(commands, args['l4t_repo'])
    elif new_driver == 'dGPU':
        install_dgpu(commands, args['l4t_repo'])

    # Print out (dry run) or run the commands to remove/install the drivers.
    if (args['dry']):
        print_commands(commands, args['verbose'])
    else:
        logger.info("=== INSTALL SUMMARY ===\n")
        print_commands(commands, verbose=False)
        logger.info("")
        run_commands(commands, args['interactive'], args['log'])
        logger.info(f"\n{new_driver} package installation is complete!")
        logger.info(f"Please plug the display cable into the {new_driver} and reboot")
        logger.info(f"the system now to complete the switch to the {new_driver}.")

if __name__ == "__main__":
    filename = os.path.basename(__file__)
    description = "Utility to switch between the NVIDIA Jetson iGPU and dGPU driver stacks."
    examples = f'''examples:
  Show the currently installed drivers:
    # {filename} query

  Show the commands and descriptions that would be run to install the dGPU drivers:
    # {filename} install dGPU -d -v

  Install the iGPU drivers and output a log to 'logfile':
    # {filename} install iGPU -l logfile'''
    ap = argparse.ArgumentParser(description=description, epilog=examples,
            formatter_class=argparse.RawDescriptionHelpFormatter)

    subparsers = ap.add_subparsers(dest='action')
    subparsers.required = True

    # Add the 'query' command.
    parser_query = subparsers.add_parser('query', help="query the current driver stack")

    # Add the 'install' command and options.
    parser_install = subparsers.add_parser('install', help="install the iGPU or dGPU driver stack")
    parser_install.add_argument('driver', nargs=1, choices=['iGPU','dGPU'],
            help="install iGPU or dGPU driver stack")
    parser_install.add_argument("-f", "--force", action='store_true',
            help="force reinstallation of the specified driver stack")
    parser_install.add_argument("-d", "--dry", action='store_true',
            help="do a dry run, showing the commands that would be executed but not actually executing them")
    parser_install.add_argument("-i", "--interactive", action='store_true',
            help="run commands interactively (asks before running each command)")
    parser_install.add_argument("-v", "--verbose", action='store_true',
            help="enable verbose output (used with --dry to describe the commands that would be run)")
    parser_install.add_argument("-l", "--log",
            help="writes a log of the install to the specified file")
    parser_install.add_argument("-r", "--l4t-repo", nargs='?', default=l4t_repo,
            help=f"specify the L4T apt repo (i.e. when using an apt mirror; default is {l4t_repo})")

    args = vars(ap.parse_args())
    if args['action'] == 'query':
        query()
    elif args['action'] == 'install':
        install(args)
